/**
*	SmallBasic byte-code executor
*
*	Nicholas Christopoulos
*
*	This program is distributed under the terms of the GPL v2.0 or later
*	Download the GNU Public License (GPL) from www.gnu.org
*/

#include "sys.h"
#include "panic.h"
#include "blib.h"
#include "device.h"
#include "str.h"

#include "kw.h"
#include "var.h"
#include "smbas.h"

#define BRUN_MODULE
#include "pproc.h"

byte	opt_graphics = 1;

#if defined(_PalmOS)
#define	idSmBa	0x536D4261
#define	idDATA	0x44415441
#endif

//#define _CHECK_STACK

word	prog_line;			// The current source line
int		prog_error;			// The last error code (its work as flag)

word	prog_length;		// The byte-code length (program length in bytes)
word	prog_ip;			// The instruction pointer
byte	*prog_source;		// The byte-code
word	data_ip;			// READ/DATA position

var_t	**tvar;				// The table of variables
lab_t	*tlab;				// The table of labels

stknode_t	*prog_stack;	// The program stack 
word	prog_stack_alloc;	// The stack size
word	prog_stack_count;	// The stack pointer
word	prog_varcount;
word	prog_labcount;
char	*prog_file;
static dword evt_check_every;

#if defined(_PalmOS) 
static char	fileName[65];
#else
static char	fileName[1024];
#endif
mem_t	bytecode_h;
#if defined(_PalmOS)
DmOpenRef	bytecode_fp;
word		bytecode_recidx;
#define	fatal(c,m)	ErrNonFatalDisplayIf((c), (m))
#endif

/**
*	run-time error
*/
void	rt_raise(const char *fmt, ...)
{
	char	*buff;
	va_list ap;

	va_start(ap, fmt);

	prog_error = 1;

	buff = tmp_alloc(256);
	#if defined(_PalmOS)
	StrVPrintF(buff, fmt, ap);
	#else
	vsprintf(buff, fmt, ap);
	#endif
	va_end(ap);
	
	dev_printf("* ERROR AT %s:%d *\n", prog_file, prog_line);
	dev_printf("> %s\n", buff);

	tmp_free(buff);
}

/* ERROR MESSAGES */
void	err_stackoverflow()		SEC(TRASH);
void	err_stackoverflow()		{	rt_raise("STACK OVERFLOW");	}

void	err_stackunderflow()	SEC(TRASH);
void	err_stackunderflow()	{	rt_raise("STACK UNDERFLOW");	}

void	err_arrmis_lp()			SEC(TRASH);
void	err_arrmis_lp()			{	rt_raise("ARRAY: MISSING '('"); }

void	err_arrmis_rp()			SEC(TRASH);
void	err_arrmis_rp()			{	rt_raise("ARRAY: MISSING ')'"); }

void	err_arridx()			SEC(TRASH);
void	err_arridx()			{	rt_raise("ARRAY: INDEX OUT OF RANGE");	}

void	err_typemismatch()		SEC(TRASH);
void	err_typemismatch()		{	rt_raise("TYPE MISMATCH");	}

void	err_varisnotarray()		SEC(TRASH);

/*
*	returns the code of the next command
byte	code_peek()
{
	return prog_source[prog_ip];
}
*/

/*
*	moves the instruction pointer to the next instruction
byte	code_getnext()
{
	byte	v;

	v = prog_source[prog_ip];
	prog_ip ++;
	return v;
}
*/

/*
*	returns the next 16bit and moves the instruction pointer to the next instruction
word	code_getnext16()
{
	word	v;

	memcpy(&v, prog_source+prog_ip, 2);
	prog_ip += 2;
	return v;
}
*/

/*
*	returns the next 32bit and moves the instruction pointer to the next instruction
*/
dword	code_getnext32()
{
	dword	v;

	memcpy(&v, prog_source+prog_ip, 4);
	prog_ip += 4;
	return v;
}

/*
*	returns the next 64bit and moves the instruction pointer to the next instruction
*/
double	code_getnext64f()
{
	double	v;

	memcpy(&v, prog_source+prog_ip, 8);
	prog_ip += 8;
	return v;
}

/*
*	jump to label
*/
void	code_jump_label(word label_id)
{
	prog_ip = tlab[label_id].ip;
}

/*
*	jump to IP
void	code_jump(word ip)
{
	prog_ip = ip;
}
*/

/*
*	Put the node 'node' in stack (PUSH)
*/
void	code_push(stknode_t *node)
{
#if defined(_UnixOS) && defined(_CHECK_STACK)
	int		i;
#endif
	if	( prog_stack_count + 1 >= prog_stack_alloc )	{
		err_stackoverflow();
		return;
		}

	prog_stack[prog_stack_count] = *node;
#if defined(_UnixOS) && defined(_CHECK_STACK)
	for ( i = 0; keyword_table[i].name[0] != '\0'; i ++ )	{
		if	( node->type == keyword_table[i].code )	{
			printf("%3d: PUSH %s (%d)\n", prog_stack_count, keyword_table[i].name, prog_line);
			break;
			}
		}
#endif
	prog_stack_count ++;
}

/*
*	Returns and deletes the topmost node from stack (POP)
*/
void	code_pop(stknode_t *node)
{
#if defined(_UnixOS) && defined(_CHECK_STACK)
	int		i;
#endif
	if	( prog_stack_count )	{
		prog_stack_count --;
		if	( node )
			*node = prog_stack[prog_stack_count];
#if defined(_UnixOS) && defined(_CHECK_STACK)
		for ( i = 0; keyword_table[i].name[0] != '\0'; i ++ )	{
			if	( prog_stack[prog_stack_count].type == keyword_table[i].code )	{
				printf("%3d: POP %s (%d)\n", prog_stack_count, keyword_table[i].name, prog_line);
				break;
				}
			}
#endif
		}
	else	{
		if	( node )	{
			err_stackunderflow();
			node->type = 0xFF;
			}
		}
}

/*
*	Peek the topmost node of stack
*/
stknode_t*	code_stackpeek()
{
	if	( prog_stack_count )
		return &prog_stack[prog_stack_count-1];
	
	return NULL;
}

/*
*	Returns the varptr of the next variable
*	if the variable is an array returns the element ptr
*/
var_t	*code_getvarptr()
{
	var_t	*basevar_p, *var_p = NULL;
	long	array_index;

	basevar_p = tvar[code_getnext16()];
	if	( basevar_p->type == V_ARRAY )	{
		/*
		*	Variable is an array
		*/
		if	( code_peek() != kwTYPE_LEVEL_BEGIN )
			err_arrmis_lp();
		else	{
			code_skipnext();	// '('
			array_index = getarrayidx(basevar_p);
			if	( !prog_error )	{
				if	( array_index < basevar_p->size )	{
					var_p = (var_t*) (basevar_p->ptr + (array_index * sizeof(var_t)));
					if	( code_peek() == kwTYPE_SEP )	
						code_skipnext();		// ','
					else if	( code_peek() == kwTYPE_LEVEL_END )	
						code_skipnext();		// ')', ')' level
					else
						err_arrmis_rp();
					}
				else	
					err_arridx();
				}
			}
		}
	else	{
		var_p = basevar_p;
		if	( code_peek() == kwTYPE_LEVEL_BEGIN )
			err_varisnotarray();
		}

	return var_p;
}

/*
*	returns true if the next element is a null-array i.e. v() keyword
*
*	example:
*	v()    --- true
*   v(1)   --- false
*/
int		code_isnullarray_nc(int start)
{
	var_t	*basevar_p;
	int16	idx;

	memcpy(&idx, prog_source+prog_ip+start, 2);
	if	( idx < prog_varcount )	{
		basevar_p = tvar[idx];
		if	( basevar_p->type == V_ARRAY )	{
			if	( prog_source[prog_ip+start+2] != kwTYPE_LEVEL_BEGIN )
				return 1;
			}
		}
	return 0;
}

int		code_isnullarray()
{
	if	( prog_source[prog_ip] == kwTYPE_VAR )	
		return code_isnullarray_nc(1);
	return 0;
}

/*
*/
void	setsysvar_int(int index, long value)
{
	var_t	*var_p = tvar[index];
	
	var_p->type = V_INT;
	var_p->const_flag = 1;
	var_p->i = value;
}

/*
*/
void	setsysvar_num(int index, double value)
{
	var_t	*var_p = tvar[index];
	
	var_p->type = V_NUM;
	var_p->const_flag = 1;
	var_p->n = value;
}

/*
*/
void	setsysvar_str(int index, const char *value)
{
	var_t	*var_p = tvar[index];
	int		l = strlen(value)+1;
	
	if	( var_p->type == V_STR )
		tmp_free(var_p->ptr);

	var_p->type = V_STR;
	var_p->const_flag = 1;
	var_p->ptr = tmp_alloc(l);
	strcpy(var_p->ptr, value);
	var_p->size = l;
}

/*
*	RUN byte-code
*	
*	source structure:
*		[word]	number_of_variables
*		[word]	number_of_labels
*		for each label:
* 			[word]	... label-ip
*		[word]	bytecode_length
*		...
*
*	brun_init(source)
*	brun()
*	brun_close()
*/
void	brun_init(byte *source)
{
	word	i;

	memcpy(&prog_varcount, source  , 2);
	memcpy(&prog_labcount, source+2, 2);

	if	( prog_varcount == 0 )
		prog_varcount ++;
	tvar = tmp_alloc(sizeof(var_t *) * prog_varcount);
	for ( i = 0; i < prog_varcount; i ++ )	
		tvar[i] = v_new();

	if	( prog_labcount == 0 )	
		tlab = tmp_alloc(sizeof(lab_t));
	else	{
		tlab = tmp_alloc(sizeof(lab_t) * prog_labcount);
		for ( i = 0; i < prog_labcount; i ++ )	
			memcpy(&tlab[i].ip, source+(i<<1)+4, 2);
		}

	prog_stack_alloc = 256;
	prog_stack = tmp_alloc(sizeof(stknode_t) * prog_stack_alloc);
	prog_stack_count = 0;

	prog_error  = 0;
	prog_line   = 0;
	
	memcpy(&data_ip, &source[(prog_labcount<<1)+4], 2);
	memcpy(&prog_length, &source[(prog_labcount<<1)+6], 2);
	prog_source = &source[(prog_labcount << 1) + 8];
	prog_ip     = 0;

	// predefined variables
	setsysvar_int(SYSVAR_OSVER, os_ver);
	#if defined(_PalmOS)
		setsysvar_str(SYSVAR_OSNAME, "PalmOS");
	#else
		#if !defined(_Win32)
			setsysvar_str(SYSVAR_OSNAME, "Unix");
		#else
			setsysvar_str(SYSVAR_OSNAME, "Win32");
		#endif
	#endif
	setsysvar_int(SYSVAR_SBVER, SB_DWORD_VER);
	setsysvar_num(SYSVAR_PI,    3.14159265358979323846);
	setsysvar_int(SYSVAR_XMAX,	160);
	setsysvar_int(SYSVAR_YMAX,	160);
	setsysvar_int(SYSVAR_BPP,	1);
	setsysvar_int(SYSVAR_TRUE,  1);
	setsysvar_int(SYSVAR_FALSE, 0);
	setsysvar_int(SYSVAR_LINECHART, 1);
	setsysvar_int(SYSVAR_BARCHART,  2);

	//
	cmd_play_reset();
}

/*
*	load & compile the source file 'fileName'
*/
void	brun_load(char *fileName)
{
	char		*ptr;
#if defined(_PalmOS)
	DmOpenRef	fp;
	LocalID		lid;
#endif

	bc_init();
	prog_file = fileName;

	/*
	*	compile - pass1
	*/
	bc_load(fileName);

	/*
	*	compile - pass2 & binary
	*/
	bytecode_h = 0;
	if	( !bc_get_error() )	{
		bc_pass2();
		if	( !bc_get_error() )	{
			bc_check_labels();
			if	( !bc_get_error() )	
				bytecode_h = bc_createbin();
			}
		}
	bc_close();

	/*
	*	run
	*/
	if	( bytecode_h )	{
		#if defined(_PalmOS)
		/*
		*	PalmOS - Memory optimization
		*
		*	We'll free 'bytecode' memory by using a database record handle
		*	Because 'bytecode' it is working as read-only we have no problems
		*	using record's memory handle instead of a dynamic-RAM handle
		*/
		word		recIndex;
		mem_t		rec_h;
		char		*prevrec, *newrec;

		lid = DmFindDatabase(0, (char *) "SBVM.BIN");
		if	( lid )	
			DmDeleteDatabase(0, lid);
		DmCreateDatabase(0, "SBVM.BIN", idSmBa, idDATA, 0);
		lid = DmFindDatabase(0, (char *) "SBVM.BIN");
		fp = DmOpenDatabase(0, lid, dmModeReadWrite);

		// save bytecode_h
		recIndex = dmMaxRecordIndex;
		rec_h = DmNewRecord(fp, &recIndex, MemHandleSize(bytecode_h));

		newrec = mem_lock(bytecode_h);
		prevrec = mem_lock(rec_h);
		DmWrite(prevrec, 0, newrec, MemHandleSize(bytecode_h));
		mem_unlock(rec_h);
		mem_unlock(bytecode_h);
		DmReleaseRecord(fp, recIndex, 1);

		// free bytecode_h
		mem_free(bytecode_h);
		bytecode_h = DmGetRecord(fp, recIndex);

		bytecode_fp = fp;
		bytecode_recidx = recIndex;

		/*
		*	Try to defrag dynamic memory
		*/
		MemHeapCompact(0);
		#endif

		ptr = mem_lock(bytecode_h);
		brun_init(ptr);
		#if defined(_PalmOS)
		EvtWakeup();		// !@$!@#$@#!$#@#$$!
		#endif
		}
	else
		prog_error = -32;
}

int		brun_close(void)
{
	word	i;

	if	( bytecode_h )	{
		// clean up
		for ( i = 0; i < prog_varcount; i ++ )	{
			v_free(tvar[i]);
			tmp_free(tvar[i]);
			}
		tmp_free(tvar);
		tmp_free(tlab);
		tmp_free(prog_stack);

		mem_unlock(bytecode_h);
		#if defined(_PalmOS)
		DmReleaseRecord(bytecode_fp, bytecode_recidx, 1);
		DmRemoveRecord(bytecode_fp, bytecode_recidx);
		DmCloseDatabase(bytecode_fp);
		#else
		mem_free(bytecode_h);
		#endif
		bytecode_h = 0;
		}

	//
	if	( prog_error != -1 && prog_error != 0 )
		return 1;
	return 0;
}

/*
*	BREAK
*/
void	brun_stop()
{
	prog_error = -3;
}

/*
*	returns the status of executor (runing or stopped)
*/										
int		brun_status()
{
	if	( prog_error )
		return BRUN_STOPPED;
	return BRUN_RUNNING;
}

/*
*	BREAK - display message, too
*/
void	brun_break()
{
	if	( brun_status() == BRUN_RUNNING )	{
		if	( os_graphics )	
			dev_settextcolor(0, 15);
		dev_printf("\n\a* BREAK AT LINE %d *\n", prog_line);
		}
	brun_stop();
}

/*
*/
void	cmd_chain(void)
{
	var_t	var;

	v_init(&var);
	eval(&var);
	if	( prog_error )	
		return;
	else	{
		if	( var.type != V_STR )	{
			err_typemismatch();
			return;
			}
		else	{
			strcpy(fileName, var.ptr);
			if	( !basFileExist(fileName) )	{
				rt_raise("CHAIN: FILE '%s' DOES NOT EXIST", fileName);
				v_free(&var);
				return;
				}
			v_free(&var);
			}
		}

	brun_close();
	#if defined(_PalmOS)
	dev_cls();
	#endif
	brun_load(fileName);
}

/*
*	RUN "program"
*
*	For PalmOS use RUN CreatorID (i.e. run "calc")
*/
void	cmd_run(void)
{
	var_t	var;

	v_init(&var);
	eval(&var);
	if	( prog_error )	
		return;
	else	{
		if	( var.type != V_STR )	{
			err_typemismatch();
			return;
			}
		else	{
			strcpy(fileName, var.ptr);
			v_free(&var);
			}
		}

//	brun_close();
	if	( !dev_run(fileName) )
		rt_raise("RUN \"%s\" FAILED", fileName);
}

/*
static void	run_events()
{
	static dword next_check;
	register dword now;

	#if defined(_PalmOS)
	now = TimGetTicks();
	#else
	now = clock();
	#endif

	// events
	if	( now >= next_check )	{
		next_check = now + evt_check_every;

		switch ( dev_events(0) )		{
		case	-1:
			break;				// break event
		case	-2:
			prog_error=-2;
			if	( os_graphics )	
				dev_settextcolor(0, 15);
			dev_printf("\n\a* BREAK AT LINE %d *\n", prog_line);
			break;
			};
		}
}
*/

/*
*/
void	bc_loop(int isf)
{
	byte	code, pops;
	byte	trace_flag = 0;
	word	next_ip;
	stknode_t	node;
	int		udf_level = 0;
	static dword next_check;
	register dword now;


	/**
	*	DO NOT CREATE FUNCTION-PTR-TABLE
	*	function tables does not works with multiseg
	*
	*	For commands witch changes the IP use
	*
	*   case mycommand:
	*		command();
	*		if	( prog_error )	break;
	*		continue;
	*/
	while ( prog_ip < prog_length )	{

//		run_events();

		#if defined(_PalmOS)
		now = TimGetTicks();
		#else
		now = clock();
		#endif

		// events
		if	( now >= next_check )	{
			next_check = now + evt_check_every;

			switch ( dev_events(0) )		{
			case	-1:
				break;				// break event
			case	-2:
				prog_error=-2;
				if	( os_graphics )	
					dev_settextcolor(0, 15);
				dev_printf("\n\a* BREAK AT LINE %d *\n", prog_line);
				break;
				};
			}

		//
		if	( !prog_error )	{
			// next command
			code = prog_source[prog_ip];
			prog_ip ++;

			switch ( code )	{
			case	kwLABEL:
			case	kwREM:
			case	kwTYPE_EOC:
				continue;
			case	kwTYPE_LINE:
				prog_line = code_getnext16();
				if	( trace_flag )
					dev_printf("<%d>", prog_line);
				continue;
			case 	kwLET:
				cmd_let(0);
				break;
			case 	kwCONST:
				cmd_let(1);
				break;
			case	kwGOTO:
				next_ip = code_getnext16();

				// clear the stack (whatever you can)
				pops = code_getnext();
				while ( pops > 0 )	{
					code_pop(NULL);
					pops --;
					}

				// jump
				prog_ip = next_ip;
				continue;
			case	kwGOSUB:
				cmd_gosub();
				if	( prog_error )	break;
				continue;
			case	kwRETURN:
				cmd_return();
				if	( prog_error )	break;
				continue;
			case	kwONJMP:
				cmd_on_go();
				if	( prog_error )	break;
				continue;
			case	kwPRINT:
				cmd_print();
				break;
			case	kwINPUT:
				cmd_input();
				break;
			case	kwCLS:
				dev_cls();
				graph_reset();
				break;
			case	kwIF:
				cmd_if();
				if	( prog_error )	break;
				continue;
			case	kwELIF:
				cmd_elif();
				if	( prog_error )	break;
				continue;
			case	kwELSE:
				cmd_else();
				if	( prog_error )	break;
				continue;
			case	kwENDIF:
				code_pop(&node);
				// 'GOTO' SHIT
				while ( node.type != kwIF )	{ code_pop(&node); if ( prog_error ) break; }
				if	( !prog_error )	{
					prog_ip += 4;
					continue;
					}
				break;
			case	kwFOR:
				cmd_for();
				if	( prog_error )	break;
				continue;
			case	kwNEXT:
				cmd_next();
				if	( prog_error )	break;
				continue;
			case	kwWHILE:
				cmd_while();
				if	( prog_error )	break;
				continue;
			case	kwWEND:
				cmd_wend();
				if	( prog_error )	break;
				continue;
			case	kwREPEAT:
				node.type = kwREPEAT;
				prog_ip += 2;
				node.exit_ip = code_getnext16();
				code_push(&node);	
				continue;
			case	kwUNTIL:
				cmd_until();
				if	( prog_error )	break;
				continue;
			case 	kwDIM:
				cmd_dim();
				break;
			case 	kwERASE:
				cmd_erase();
				break;
			case	kwREAD:
				cmd_read();
				break;
			case	kwDATA:
				cmd_data();
				break;
			case	kwRESTORE:
				cmd_restore();
				break;

			case	kwTYPE_CALL_UDP:		// user defined procedure
				cmd_udp(kwPROC);
				if	( isf )	
					udf_level ++;
				if	( prog_error )	break;
				continue;
			case	kwTYPE_CALL_UDF:		// user defined function
				if	( isf )	{
					cmd_udp(kwFUNC);
					udf_level ++;
					}
				else
					err_syntax();
				if	( prog_error )	break;
				continue;
			case	kwTYPE_RET:
				cmd_udpret();
				if	( isf )	{
					udf_level --;
					if	( udf_level == 0 )
						return;
					}
				if	( prog_error )	break;
				continue;
			case	kwTYPE_CRVAR:
				cmd_crvar();
				break;
			case	kwTYPE_PARAM:
				cmd_param();
				break;
			case	kwEXIT:
				cmd_exit();
				if	( prog_error )	break;
				continue;

			// ...
			case	kwPEN:
				cmd_pen();
				break;
			case	kwRANDOMIZE:
				cmd_randomize();
				break;

			// graphics
			case	kwPSET:
				cmd_pset();
				break;
			case	kwLINE:
				cmd_line();
				break;
			case	kwRECT:
				cmd_rect();
				break;
			case	kwCIRCLE:
				cmd_circle();
				break;
			case	kwDRAWPOLY:
				cmd_drawpoly();
				break;


			// third class
			case	kwAT:
				cmd_at();
				break;
			case	kwLOCATE:
				cmd_locate();
				break;
			case	kwCOLOR:
				cmd_color();
				break;
			case	kwBEEP:
				cmd_beep();
				break;
			case	kwSOUND:
				cmd_sound();
				break;
			case	kwPAUSE:
				cmd_pause();
				break;
			case	kwDELAY:
				cmd_delay();
				break;
			case	kwARC:
				cmd_arc();
				break;
			case	kwDRAW:
				cmd_draw();
				break;
			case	kwPLAY:
				cmd_play();
				break;

			// --- at end ---
			case	kwWSPLIT:
				cmd_wsplit();
				break;
			case	kwOPEN:
				cmd_fopen();
				break;
			case	kwCLOSE:
				cmd_fclose();
				break;
			case	kwFILEPRINT:
				cmd_fprint();
				break;
			case	kwLINEINPUT:
				cmd_flineinput();
				break;
			case	kwFILEINPUT:
				cmd_finput();
				break;
			case	kwSEEK:
				cmd_fseek();
				break;
			case	kwCOPY:
				cmd_filecp(0);
				break;
			case	kwRENAME:
				cmd_filecp(1);
				break;
			case	kwKILL:
				cmd_fkill();
				break;
			case	kwTRON:
				trace_flag = 1;
				continue;
			case	kwTROFF:
				trace_flag = 0;
				continue;

			case	kwCHART:
				cmd_chart();
				break;

			case	kwSTOP:
			case	kwEND:
				prog_error = -1;
				break;
			case	kwCHAIN:
				cmd_chain();
				break;

			case	kwRUN:
				cmd_run();
				break;

			case	kwLOGPRINT:
				cmd_logprint();
				break;
	#ifdef BC_DEBUG
	/* This debugging command will dump the next few bytes of the byte code
	** program then exit.
	*/
			case kwBCDUMP:
				cmd_bcdump();
				prog_error = -1;
				break;

			case	kwSTKDUMP:
				dev_print("\nSTKDUMP:\n");
				dump_stack();
				prog_error = -1;
				break;
	#endif

			default:
				rt_raise("OUT OF CODE SEG->CODE[0x%2X]=0x%X", prog_ip, prog_source[prog_ip]);
				}
			}

		// too many parameters
		code = prog_source[prog_ip];
		if	( code != kwTYPE_EOC && code != kwTYPE_LINE && !prog_error )
			rt_raise("PARAM COUNT ERROR @%d=%X", prog_ip, prog_source[prog_ip]);
		else	{
			prog_ip ++;
			if ( code == kwTYPE_LINE )	{
				prog_line = code_getnext16();
				if	( trace_flag )
					dev_printf("<%d>", prog_line);
				}
			}

		// quit on error
		if	( prog_error )	{
			// cleanup stack
			while ( prog_stack_count > 0 )	{
				code_pop(&node);
				switch ( node.type )	{
				case	kwTYPE_CRVAR:
					v_free(tvar[node.x.vdvar.vid]);					// free local variable data
					tmp_free(tvar[node.x.vdvar.vid]);
					tvar[node.x.vdvar.vid] = node.x.vdvar.vptr;		// restore ptr
					break;
				case	kwBYREF:
					tvar[node.x.vdvar.vid] = node.x.vdvar.vptr;
					break;
				case	kwTYPE_VAR:
					if	( (node.x.param.vcheck == 1) || (node.x.param.vcheck == 0x81) ) {
						v_free(node.x.param.res);
						tmp_free(node.x.param.res);
						}
						
					break;
				case	kwTYPE_RET:
					v_free(node.x.vdvar.vptr);				// free ret-var
					tmp_free(node.x.vdvar.vptr);
					break;
				case	kwFUNC:
					tvar[node.x.vcall.rvid] = node.x.vcall.retvar;	// restore ptr
					break;
					};
				}
			break;
			}

		}
}

/*
*	executor
*/
void	brun(char *file)
{
	dword	tps;

	brun_load(file);
	if	( prog_error )
		return;

	dev_init(opt_graphics,0);

	#if defined(_PalmOS)
	tps = SysTicksPerSecond();
	#else
	tps = CLOCKS_PER_SEC;
	#endif
	evt_check_every = (50 * tps) / 1000;	// 50ms

	bc_loop(0);

	if	( prog_error == -1 || prog_error == 0 )	{	// normal exit
		if	( os_graphics )	
			dev_settextcolor(0, 15);

		dev_print("\n\a* DONE *\n");
		prog_error = -1;
		}

	brun_close();
	
	dev_restore();
}

/*
*	debug info
*	stack dump
*/
void	dump_stack()
{
	stknode_t	node;
	int			i;

	do	{
		code_pop(&node);
		if	( node.type != 0xFF )	{
			for ( i = 0; keyword_table[i].name[0] != '\0'; i ++ )	{
				if	( node.type == keyword_table[i].code )	{
					dev_printf("\t%s", keyword_table[i].name);
					switch ( node.type )	{
					case kwGOSUB:
						dev_printf(" RIP: %d", node.x.vgosub.ret_ip);
						if	( prog_source[node.x.vgosub.ret_ip] == kwTYPE_LINE )	{
							dev_printf(" = LI %d", (*((word *) (prog_source+node.x.vgosub.ret_ip+1))) - 1);
							}
						break;
						}
					dev_print("\n");
					break;
					}
				}
			}
		else
			break;
		} while ( 1 );
}

#if !defined(_PalmOS) && !defined(_BCB_W32_IDE)
//////////////////////////////////////////////////////////////////////////////////////
//////// Command-line interface (Unix)

/*
*	decompiler
*/
void	dump_bytecode()
{
	int		c, b, d, h, l, j;
	int		len, new_ip;

	fprintf(stderr, "Program size=%d\n", prog_length);
	do	{
		b = code_getnext();
		h = b >> 4;
		l = b & 0xF;

		printf("%06d: %c%c %c ",
			(int) prog_ip-1,
			"0123456789ABCDEF"[h],
			"0123456789ABCDEF"[l],
			(b>=32) ? b : 42 );

		switch ( b )	{
		case	kwTYPE_LINE:
			printf("--LINE: %d", code_getnext16());
			break;
		case	kwTYPE_EVPUSH:
			printf("PUSH RESULT");
			break;
		case	kwTYPE_EVPOP:
			printf("POP LEFT");
			break;
		case	kwTYPE_EOC:
			printf("--EOC");
			break;
		case	kwPROC:
			printf("=== PROCEDURE ===");
			break;
		case	kwFUNC:
			printf("=== FUNCTION ===");
			break;
		case	kwFILEPRINT:
			printf("FILE PRINT");
			break;
		case	kwFILEINPUT:
			printf("FILE INPUT");
			break;
		case	kwLINEINPUT:
			printf("LINE INPUT");
			break;
		case	kwFORSEP:
			printf("FOR SEP");
			break;
		case	kwINPUTSEP:
			printf("INPUT SEP");
			break;
		case	kwTYPE_PARAM:
		case	kwTYPE_CRVAR:
			c = code_getnext();
			if	( b == kwTYPE_PARAM )
				printf("PARAM = %d: ", c);
			else
				printf("CRVAR = %d: ", c);
				
			d = 0;
			for ( j = 0; j < c; j ++ )	{
				if	( b == kwTYPE_PARAM )	{
					d = code_getnext();
					if	( d & 0x80 )
						printf("&");
					}
				printf("%d ", code_getnext16());
				if	( d & 0x1 )
					printf("()");
				}
			break;
		case	kwTYPE_RET:
			printf("=== UDP/F RET ===");
			break;
		case	kwTYPE_CALL_UDP:
			printf("CALL UDP %d", code_getnext16());
			code_skipnext16();
			break;
		case	kwTYPE_CALL_UDF:
			printf("CALL UDF %d", code_getnext16());
			printf(", RVID: %d", code_getnext16());
			break;
		case	kwEXIT:
			printf("EXIT ");
			c = code_peek();
			if	( c == kwFOR )	{ printf("FOR"); code_skipnext(); }
			if	( c == kwLOOP )	{ printf("LOOP"); code_skipnext(); }
			if	( c == kwPROC )	{ printf("PROC"); code_skipnext(); }
			if	( c == kwFUNC )	{ printf("FUNC"); code_skipnext(); }
			break;

		default:
			for ( c = 0; keyword_table[c].name[0] != '\0'; c ++)	{
				if ( b == keyword_table[c].code )	{
					printf("%s ", keyword_table[c].name);
					break;
					}
				}

			switch ( b )	{
			case	kwTYPE_SEP:
				c = code_getnext();
				printf("SEPARATOR \'%c\' ", (c>=32) ? c : '?');
				break;
			case	kwTYPE_LOGOPR:
			case	kwTYPE_CMPOPR:
			case	kwTYPE_ADDOPR:
			case	kwTYPE_MULOPR:
			case	kwTYPE_POWOPR:
			case	kwTYPE_UNROPR:
				c = code_getnext();
				printf("DATA %d \'%c\' ", c, (c>=32) ? c : '?');
				break;
			case	kwTYPE_LEVEL_BEGIN:
				printf("(");
				break;
			case	kwTYPE_LEVEL_END:
				printf(")");
				break;
			case	kwONJMP:
				new_ip = code_getnext16();
				code_skipnext16();
				b = code_getnext();
				c = code_getnext();
				printf("ON %s CNT %d\n", ((b == kwGOTO)?"GOTO":"GOSUB"), c);
				printf("      NEXT IP %d\n", new_ip);
				for ( j = 0; j < c; j ++ )	
					printf("       JMP IP %d\n", code_getnext16());
				printf("ON %s EXPR:", ((b == kwGOTO)?"GOTO":"GOSUB"));
				break;
			case	kwGOTO:
				new_ip = code_getnext16();
				printf("= JMP %d ", new_ip);
				c = code_getnext();
				printf("POPs %d", c);
				break;
			case	kwGOSUB:
				c = code_getnext16();
				printf("ID %d ", c);
				new_ip = tlab[c].ip;
				printf("= JMP %d ", new_ip);
				break;
			case	kwRESTORE:
			case	kwTYPE_LINE:
			case	kwLABEL:
			case	kwTYPE_VAR:
			case	kwLET:
				printf("ID %d ", code_getnext16());
				break;
			case	kwTYPE_INT:
				printf("VAL %d ", code_getnext32());
				break;
			case	kwTYPE_NUM:
				printf("VAL %f ", code_getnext64f());
				break;
			case	kwTYPE_STR:
				len = code_getnext16();
				printf("\"");
				for ( j = 0; j < len; j ++ )
					printf("%c", prog_source[prog_ip+j]);
				printf("\"");
				prog_ip += len;	
				break;
			case	kwIF:		case	kwFOR: 		case	kwWHILE:	case	kwREPEAT:
			case	kwELSE:		case	kwELIF:
			case	kwENDIF:	case	kwNEXT:		case	kwWEND:		case	kwUNTIL:
				printf("IP1 %d IP2 %d ", code_getnext16(), code_getnext16());
				break;
				}
			}

		printf("\n");
		} while	( prog_ip < prog_length );
}

/*
*	main
*/
int	main(int argc, char *argv[])
{
	char	file[1024];
	char	prev_cwd[1024];
	char	cwd[1024], *slash;
	byte	opt_decomp = 0; //, opt_graphics = 0;
	int		i, use_stdin = 0, c;
	FILE	*fp;

	printf("SmallBASIC byte-code executor version %s\n", SB_STR_VER);
	printf("Written by Nicholas Christopoulos\n");
	strcpy(file, "");
	opt_graphics = 0;
	for ( i = 1; i < argc; i ++ )	{
		if	( argv[i][0] == '-' )	{
			if	( argv[i][1] == 's' )
				opt_decomp ++;
			if	( argv[i][1] == 'g' )
				opt_graphics = 2;
			if	( argv[i][1] == 'j' )
				os_charset = enc_sjis;
			if	( argv[i][1] == 'b' )
				os_charset = enc_big5;
			if	( argv[i][1] == 'm' )
				os_charset = enc_gmb;
			if	( argv[i][1] == 'u' )
				os_charset = enc_unicode;
			if	( argv[i][1] == 'h' )	{
				fprintf(stderr, "usage: sbrun source [options]\n");
				fprintf(stderr, "-s		decompiler\n");
				fprintf(stderr, "-g		enable graphics\n");
				fprintf(stderr, "\ncharset (default: utf8)\n");
				fprintf(stderr, "-j		enable sjis\n");
				fprintf(stderr, "-b		enable big5\n");
				fprintf(stderr, "-m		enable generic multibyte\n");
				fprintf(stderr, "-u		enable unicode!\n");
				return 1;
				}
			}
		else
			strcpy(file, argv[i]);
		}

	//
	getcwd(prev_cwd, 1024);
	strcpy(cwd, prev_cwd);

	if	( strlen(file) == 0 )	{
		// stdin
		use_stdin ++;
		#if defined(_Win32)
		sprintf(file, "sbrun.tmp");
		#else
		sprintf(file, "/tmp/sb%d.bas", getpid());
		#endif

		fp = fopen(file, "wb");
		if	( fp )	{
			while ( (c = fgetc(stdin)) != -1 )	
				fputc(c, fp);
			fclose(fp);
			}
		else
			perror("sbrun");
		}
	else	{
		#if defined(_Win32)
		slash = strrchr(file, '\\');
		#else
		slash = strrchr(file, '/');
		#endif
		if	( slash )	{
			char	tmp[1024];
			*slash = '\0';
			
			#if defined(_Win32)
			strcat(cwd, "\\");
			#else
			strcat(cwd, "/");
			#endif
			strcat(cwd, file);
			strcpy(tmp, slash+1);
			strcpy(file, tmp);

//			printf("Current directory changed to [%s]\n", cwd);
//			printf("Source file changed to: [%s]\n", file);
			chdir(cwd);
			}
		}

	if	( opt_decomp )	{
		brun_load(file);
		if	( bc_get_error() )
			return 1;
		dump_bytecode();
		}
	else	{
		// run
		memmgr_init();
		brun(file);
		}

	// cleanup
	if	( use_stdin )
		remove(file);
	chdir(prev_cwd);
	return 0;
}
#endif

